#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest
import re

kernel_trace_list = sorted(["addition_kernel", "subtract_kernel"])
kernel_counter_list = ["addition_kernel"]


def unique(lst):
    return list(set(lst))


def test_counter_collection_json_data(json_data):
    data = json_data["rocprofiler-sdk-tool"]

    def get_kernel_name(kernel_id):
        return data["kernel_symbols"][kernel_id]["formatted_kernel_name"]

    counter_collection_data = data["callback_records"]["counter_collection"]

    for counter in counter_collection_data:
        kernel_name = get_kernel_name(counter.dispatch_data.dispatch_info.kernel_id)
        assert kernel_name in kernel_counter_list


def test_kernel_trace_json(json_data):
    data = json_data["rocprofiler-sdk-tool"]

    def get_kernel_name(kernel_id):
        return data["kernel_symbols"][kernel_id]["formatted_kernel_name"]

    def get_kind_name(kind_id):
        return data["strings"]["buffer_records"][kind_id]["kind"]

    kernel_dispatch_data = data["buffer_records"]["kernel_dispatch"]
    kernels = []
    assert len(kernel_dispatch_data) == 2
    for dispatch in kernel_dispatch_data:
        dispatch_info = dispatch["dispatch_info"]
        kernel_name = get_kernel_name(dispatch_info["kernel_id"])

        assert get_kind_name(dispatch["kind"]) == "KERNEL_DISPATCH"
        assert dispatch["correlation_id"]["internal"] > 0
        assert dispatch_info["agent_id"]["handle"] > 0
        assert dispatch_info["queue_id"]["handle"] > 0
        assert dispatch_info["kernel_id"] > 0
        if not re.search(r"__amd_rocclr_.*", kernel_name):
            kernels.append(kernel_name)

        assert dispatch_info["workgroup_size"]["x"] == 64
        assert dispatch_info["workgroup_size"]["y"] == 1
        assert dispatch_info["workgroup_size"]["z"] == 1
        assert dispatch_info["grid_size"]["x"] == 1024
        assert dispatch_info["grid_size"]["y"] == 1024
        assert dispatch_info["grid_size"]["z"] == 1
        assert dispatch["end_timestamp"] >= dispatch["start_timestamp"]

    assert kernels == kernel_trace_list


def test_hip_api_trace_json(json_data):
    data = json_data["rocprofiler-sdk-tool"]

    def get_operation_name(kind_id, op_id):
        return data["strings"]["buffer_records"][kind_id]["operations"][op_id]

    def get_kind_name(kind_id):
        return data["strings"]["buffer_records"][kind_id]["kind"]

    valid_domain_names = (
        "HIP_RUNTIME_API",
        "HIP_RUNTIME_API_EXT",
    )

    hip_api_data = data["buffer_records"]["hip_api"]

    functions = []
    for api in hip_api_data:
        kind = get_kind_name(api["kind"])
        assert kind in valid_domain_names
        assert api["end_timestamp"] >= api["start_timestamp"]
        functions.append(get_operation_name(api["kind"], api["operation"]))

    expected_functions = (
        [
            "hipGetDeviceCount",
            "hipSetDevice",
            "hipDeviceSynchronize",
            "hipStreamCreateWithFlags",
        ]
        + (["hipHostMalloc"] * 3)
        + (["hipMallocAsync"] * 3)
        + (["hipMemcpyAsync"] * 2)
        + [
            "hipStreamSynchronize",
            "hipDeviceSynchronize",
            "hipLaunchKernel",
            "hipGetLastError",
            "hipLaunchKernel",
            "hipGetLastError",
        ]
    )
    assert functions == expected_functions


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2023-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import pandas as pd
import argparse
import logging


def read_csv(file_path):
    df = pd.DataFrame()
    try:
        df = pd.read_csv(file_path)
    except Exception as e:
        logging.info(f"Error reading {file_path}: {e}")
        raise
    return df


def get_counter_collection_files(root_path):
    file_paths = []
    for root, _, files in os.walk(root_path):
        if "pmc_" in root:
            for file in files:
                if file.endswith("counter_collection.csv"):
                    file_path = os.path.join(root, file)
                    file_paths.append(file_path)
    return file_paths


def get_combined_df(args):
    files_list = []
    for input in args.input:
        if os.path.isfile(input):
            files_list.append(input)
        elif os.path.isdir(input):
            files_list.extend(get_counter_collection_files(input))
    if not files_list:
        raise ValueError("Valid Input files not found")
    logging.info(f"Processing files: {files_list}")
    combined_df = pd.DataFrame()
    for file in files_list:
        combined_df = pd.concat([combined_df, read_csv(file)], ignore_index=True)
    return combined_df


def write_to_file(df, args):
    logging.info(f"Saving output file to : {args.output}")
    directory, file_path = os.path.split(args.output)
    if directory:
        os.makedirs(directory, exist_ok=True)

    if not args.retain_agent_prefix:
        df["Agent_Id"] = df["Agent_Id"].apply(lambda x: x.split(" ")[-1])

    df.to_csv(args.output, index=False)


def main(args):
    logging.basicConfig(level=args.loglevel)
    input_df = get_combined_df(args)
    # Validate
    columns = [
        "Correlation_Id",
        "Dispatch_Id",
        "Agent_Id",
        "Queue_Id",
        "Process_Id",
        "Thread_Id",
        "Grid_Size",
        "Kernel_Id",
        "Kernel_Name",
        "Workgroup_Size",
        "LDS_Block_Size",
        "Scratch_Size",
        "VGPR_Count",
        "SGPR_Count",
        "Counter_Name",
        "Counter_Value",
        "Start_Timestamp",
        "End_Timestamp",
    ]
    for col in input_df.columns:
        if col not in columns:
            logging.debug(f"Unexpected column {col} found in rocprofv3 input file")

    # Convert
    indexes = [
        "Dispatch_Id",
        "Agent_Id",
        "Grid_Size",
        "Kernel_Name",
        "LDS_Block_Size",
        "Queue_Id",
        "SGPR_Count",
        "Scratch_Size",
        "VGPR_Count",
        "Workgroup_Size",
    ]

    # Drop duplicate counters in multiple PMC lines
    input_df.drop_duplicates(
        subset=indexes + ["Counter_Name"], keep="first", inplace=True
    )

    pivoted_data = input_df.pivot_table(
        index=indexes, columns="Counter_Name", values="Counter_Value", aggfunc="sum"
    ).reset_index()

    # Save
    write_to_file(pivoted_data, args)


def strtobool(val):
    """Convert a string representation of truth to true or false.
    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    if isinstance(val, (list, tuple)):
        if len(val) > 1:
            val_type = type(val).__name__
            raise ValueError(f"invalid truth value {val} (type={val_type})")
        else:
            val = val[0]

    if isinstance(val, bool):
        return val
    elif isinstance(val, str) and val.lower() in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif isinstance(val, str) and val.lower() in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        val_type = type(val).__name__
        raise ValueError(f"invalid truth value {val} (type={val_type})")


class booleanArgAction(argparse.Action):
    def __call__(self, parser, args, value, option_string=None):
        setattr(args, self.dest, strtobool(value))


def add_parser_bool_argument(gparser, *args, **kwargs):
    gparser.add_argument(
        *args,
        **kwargs,
        action=booleanArgAction,
        nargs="?",
        const=True,
        type=str,
        required=False,
        metavar="BOOL",
        default=False,
    )


def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        help="Rocprofv3 Counter Collection input files and/or directories containing `*counter_collection.csv` files",
        nargs="+",
        default=[],
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Rocprofv1 formatted output file name",
        default=None,
        type=str,
        required=True,
    )
    parser.add_argument(
        "-d",
        "--debug",
        help="Debug Logs",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.WARNING,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Verbose Logs",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
    )
    advanced_options = parser.add_argument_group("Advanced options")
    add_parser_bool_argument(
        advanced_options,
        "--retain-agent-prefix",
        help="retains the agent prefix",
    )

    return parser.parse_args()


if __name__ == "__main__":
    main(parse_args())



#!/usr/bin/env python3
###############################################################################
# MIT License
#
# Copyright (c) 2023 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

#
#   Utility Class to create the rocpd schema on an existing sqlite connection
#
#       Requires a current copy of the schema in the 'schema' subdirectory
#       Executes the contained sql 'scripts' to create the schema
#

import argparse
from . import libpyrocpd


class RocpdSchema:

    def __init__(self, uuid="", guid=""):

        variables = libpyrocpd.schema_jinja_variables()
        variables.uuid = f"{uuid}"
        variables.guid = f"{guid}"

        self.tables = RocpdSchema.load_schema(
            libpyrocpd.sql_engine.sqlite3,
            libpyrocpd.sql_schema.rocpd_tables,
            libpyrocpd.sql_option.sqlite3_pragma_foreign_keys,
            variables,
        )

        self.indexes = RocpdSchema.load_schema(
            libpyrocpd.sql_engine.sqlite3,
            libpyrocpd.sql_schema.rocpd_indexes,
            libpyrocpd.sql_option.none,
            variables,
        )

        _views = []
        for itr in ["rocpd", "data", "summary", "marker"]:
            _views += [
                RocpdSchema.load_schema(
                    libpyrocpd.sql_engine.sqlite3,
                    getattr(libpyrocpd.sql_schema, f"{itr}_views"),
                    libpyrocpd.sql_option.none,
                    variables,
                )
            ]
        self.views = "\n".join(_views)

    def write_schema(self, connection):
        connection.executescript(self.tables)
        connection.executescript(self.indexes)
        connection.executescript(self.views)

    @staticmethod
    def load_schema(engine, kind, options, variables=None, **kwargs):

        if variables is None:
            variables = libpyrocpd.schema_jinja_variables()

        for itr in ["uuid", "guid"]:
            _variable = kwargs.get(itr, None)
            if _variable is not None:
                setattr(variables, itr, f"{_variable}")

        return libpyrocpd.load_schema(engine, kind, options, variables)


def main(create=None):
    schema = RocpdSchema()

    if create:
        from . import connect

        print(f"Creating empty rpd: {args.create}")
        connection = connect(args.create)
        schema.write_schema(connection)
        connection.commit()
    else:
        print(schema.tables)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="convert rocprofiler output to an RPD database"
    )
    parser.add_argument(
        "--create", type=str, help="filename in create empty db", default=None
    )
    args = parser.parse_args()

    main(args.create)



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest
import numpy as np
import pandas as pd
import re

kernel_list = sorted(
    ["addition_kernel", "subtract_kernel", "multiply_kernel", "divide_kernel"]
)


def unique(lst):
    return list(set(lst))


def test_validate_counter_collection_pmc1_extra_counters(input_data: pd.DataFrame):
    df = input_data

    assert not df.empty
    df_agent_id = df["Agent_Id"].str.split(" ").str[-1]
    assert (df_agent_id.astype(int).values >= 0).all()
    assert (df["Queue_Id"].astype(int).values > 0).all()
    assert (df["Process_Id"].astype(int).values > 0).all()
    assert len(df["Kernel_Name"]) > 0

    counter_collection_pmc1_kernel_list = [
        x
        for x in sorted(df["Kernel_Name"].unique().tolist())
        if not re.search(r"__amd_rocclr_.*", x)
    ]

    assert kernel_list == counter_collection_pmc1_kernel_list

    kernel_count = dict([[itr, 0] for itr in kernel_list])
    assert len(kernel_count) == len(kernel_list)
    for itr in df["Kernel_Name"]:
        if re.search(r"__amd_rocclr_.*", itr):
            continue
        kernel_count[itr] += 1
    kn_cnt = [itr for _, itr in kernel_count.items()]
    assert min(kn_cnt) == max(kn_cnt) and len(unique(kn_cnt)) == 1

    assert len(df["Counter_Value"]) > 0
    assert df["Counter_Name"].str.contains("TEST_YAML_LOAD").all()
    assert (df["Counter_Value"].astype(int).values > 0).all()

    di_list = df["Dispatch_Id"].astype(int).values.tolist()
    di_uniq = sorted(df["Dispatch_Id"].unique().tolist())
    # make sure the dispatch ids are unique and ordered
    di_expect = [idx + 1 for idx in range(len(di_list))]
    assert di_expect == di_uniq


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest


# helper function
def node_exists(name, data, min_len=1):
    assert name in data
    assert data[name] is not None
    assert len(data[name]) >= min_len


def test_data_structure(input_data):
    """verify minimum amount of expected data is present"""
    node_exists("rocprofiler-sdk-json-tool", input_data)
    rocp_data = input_data
    node_exists("names", rocp_data["rocprofiler-sdk-json-tool"]["buffer_records"])
    node_exists(
        "counter_collection", rocp_data["rocprofiler-sdk-json-tool"]["buffer_records"]
    )


def test_counter_values(input_data):
    data = input_data["rocprofiler-sdk-json-tool"]
    agent_data = data["agents"]
    counter_info = data["counter_info"]
    counter_data = data["buffer_records"]["counter_collection"]

    for itr in counter_info:
        if itr["is_constant"] == 1:
            continue
        assert itr["id"]["handle"] > 0, f"{itr}"
        assert itr["is_constant"] in (0, 1), f"{itr}"
        assert itr["is_derived"] in (0, 1), f"{itr}"
        assert len(itr["name"]) >= 4, f"{itr}"
        assert len(itr["description"]) >= 4, f"{itr}"
        if itr["is_constant"] == 0:
            if itr["is_derived"] == 0:
                assert len(itr["block"]) > 0, f"{itr}"
            if itr["is_derived"] == 1:
                assert len(itr["expression"]) > 0, f"{itr}"

    def get_agent(agent_id):
        for itr in agent_data:
            if itr["id"]["handle"] == agent_id["handle"]:
                return itr
        return None

    def get_scaling_factor(agent_id):
        agent = get_agent(agent_id)
        assert agent is not None, f"id={agent_id}"
        if agent["type"] == 2 and agent["wave_front_size"] > 0:
            return 64 / agent["wave_front_size"]
        return 0

    for itr in counter_data:
        assert itr["num_records"] == len(itr["records"]), f"itr={itr}"
        agent_id = itr["dispatch_info"]["agent_id"]
        agent = get_agent(agent_id)
        scaling_factor = get_scaling_factor(agent_id)
        assert agent is not None, f"itr={itr}\nagent={agent}"
        assert itr["start_timestamp"] < itr["end_timestamp"]
        for ritr in itr["records"]:
            value = ritr["counter_value"]
            if int(round(value, 0)) > 0:
                assert int(round(value, 0)) == int(
                    round(1 * scaling_factor, 0)
                ), f"itr={itr}\nagent={agent}"


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



###############################################################################
# MIT License
#
# Copyright (c) 2023 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
###############################################################################

import sys
import os
import argparse

try:
    import ctypes

    sqlite3lib = ctypes.CDLL("libsqlite3.so")
except Exception:
    pass

from . import libpyrocpd


__all__ = ["format_path", "output_config", "add_args", "process_args"]


def _generate_attribute_docs(data):
    properties = []
    for key, itr in data.items():
        if not isinstance(key, str):
            pass
        if key.startswith("__") and key.endswith("__"):
            pass
        else:
            properties += [key]
    return "\n\t- ".join(properties)


class output_config(libpyrocpd.output_config):
    __doc__ = f"""Output configuration

    Read/Write properties:\n\t- {_generate_attribute_docs(libpyrocpd.output_config.__dict__)}

    Example:
        # folder for output data
        output_dir = os.path.join(os.getcwd(), "rocpd-output")

        # create output config instance
        cfg = output_config(output_path=output_dir, output_file="out")

        # using read/write properties
        if cfg.output_path != output_dir:
            cfg.output_path = output_dir
    """

    def __init__(self, **kwargs):
        super(output_config, self).__init__()
        self.update(**kwargs)

    def update(self, **kwargs):
        _strict = kwargs.get("strict", True)
        # _verbose = kwargs.get("log-level", "config")
        for key, itr in kwargs.items():
            if hasattr(self, key):
                # if _verbose in ("info", "trace", "config"):
                #     print(f"  - output_config.{key} = {itr}")
                if key == "agent_index_value":
                    if itr == "absolute":
                        setattr(self, key, libpyrocpd.agent_indexing.node)
                    elif itr == "type-relative":
                        setattr(self, key, libpyrocpd.agent_indexing.logical_node_type)
                    else:
                        setattr(self, key, libpyrocpd.agent_indexing.logical_node)
                else:
                    setattr(self, key, itr)
            elif _strict:
                raise KeyError(f"output_config does not have {key} attribute")
        return self


def format_path(path, tag=os.path.basename(sys.executable)):
    return libpyrocpd.format_path(path, tag)


def check_file_exists(filename):
    if not os.path.exists(filename):
        raise argparse.ArgumentTypeError(f"File '{filename}' does not exist.")
    return filename


def add_args(parser):
    """Add output arguments to an existing parser."""

    io_options = parser.add_argument_group("I/O options")

    io_options.add_argument(
        "-o",
        "--output-file",
        help="Sets the base output file name (default base filename: `out`)",
        default=os.environ.get("ROCPD_OUTPUT_NAME", "out"),
        type=str,
        required=False,
    )
    io_options.add_argument(
        "-d",
        "--output-path",
        help="Sets the output path where the output files will be saved (default path: `./rocpd-output-data`)",
        default=os.environ.get("ROCPD_OUTPUT_PATH", "./rocpd-output-data"),
        type=str,
        required=False,
    )

    kernel_naming_options = parser.add_argument_group("Kernel naming options")

    kernel_naming_options.add_argument(
        "--kernel-rename",
        help="Use ROCTx marker names instead of kernel names",
        action="store_true",
        default=False,
    )

    return ["output_file", "output_path", "kernel_rename"]


def process_args(args, valid_args):

    ret = {}
    for itr in valid_args:
        if hasattr(args, itr):
            val = getattr(args, itr)
            if itr == "output_format":
                ret[itr] = val
            elif itr == "output_path" and val is not None:
                ret[itr] = format_path(val)
            elif val is not None:
                ret[itr] = val
    return ret


def add_generic_args(parser):
    """Add generic arguments that apply to multiple output formats."""

    generic_options = parser.add_argument_group("Generic options")

    generic_options.add_argument(
        "--agent-index-value",
        choices=("absolute", "relative", "type-relative"),
        help="""Device identification format in CSV/Perfetto/OTF2 output (default: relative):
        absolute: uses node_id (Agent-0, Agent-2, Agent-4) ignoring cgroups restrictions.
        relative: uses logical_node_id (Agent-0, Agent-1, Agent-2) considering cgroups restrictions.
        type-relative: uses logical_node_type_id (CPU-0, GPU-0, GPU-1) with numbering that resets for each device type.""",
        default="relative",
    )

    return [
        "agent_index_value",
    ]


def process_generic_args(args, valid_args):
    ret = {}
    for itr in valid_args:
        if hasattr(args, itr):
            val = getattr(args, itr)
            if val is not None:
                ret[itr] = val
    return ret



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import sys
import pytest
import re
import os

from collections import Counter


def test_agent_info(agent_info_input_data):
    logical_node_id = max([int(itr["Logical_Node_Id"]) for itr in agent_info_input_data])

    assert logical_node_id + 1 == len(agent_info_input_data)

    for row in agent_info_input_data:
        agent_type = row["Agent_Type"]
        assert agent_type in ("CPU", "GPU")
        if agent_type == "CPU":
            assert int(row["Cpu_Cores_Count"]) > 0
            assert int(row["Simd_Count"]) == 0
            assert int(row["Max_Waves_Per_Simd"]) == 0
        else:
            assert int(row["Cpu_Cores_Count"]) == 0
            assert int(row["Simd_Count"]) > 0
            assert int(row["Max_Waves_Per_Simd"]) > 0


def extract_number(pattern, string):
    match = re.match(pattern, string)
    if match:
        return int(match.group(1))
    else:
        raise ValueError(f"Pattern '{pattern}' not found in '{string}'.")


def find_key_with_substring(data, substring):
    return next((k for k in data.keys() if substring in k), None)


def check_tot_data(tot_data):
    iteration_msg = find_key_with_substring(tot_data, "iterations:")
    assert tot_data[iteration_msg] == 1

    num_iterations = extract_number(r"iterations: (\d+)", iteration_msg)
    assert tot_data["main loop"] == num_iterations

    if num_iterations > 0:
        run_msg = find_key_with_substring(tot_data, "run")
        if run_msg is not None:
            value = extract_number(r"run\((\d+)\)", run_msg)

            assert tot_data[f"run({value})"] == num_iterations

            assert "fib" in tot_data.keys()
            assert tot_data[f"fib(n={value})"] == num_iterations
            for n in range(value, int(max([value / 2, value - 10]))):
                assert f"fib(n={n})" in tot_data.keys()

            assert tot_data[f"inefficient({value})"] == num_iterations
            assert tot_data[f"sum"] == num_iterations
            sum_msg = find_key_with_substring(tot_data, "sum(nelem=")
            assert tot_data[sum_msg] == num_iterations


def test_marker_api_trace(marker_input_data):
    functions = []

    for row in marker_input_data:
        assert row["Domain"] in [
            "MARKER_CORE_API",
            "MARKER_CORE_RANGE_API",
            "MARKER_CONTROL_API",
            "MARKER_NAME_API",
        ]
        assert int(row["Process_Id"]) > 0
        assert int(row["Thread_Id"]) == 0 or int(row["Thread_Id"]) >= int(
            row["Process_Id"]
        )
        assert int(row["End_Timestamp"]) >= int(row["Start_Timestamp"])

        functions.append(row["Function"])

    check_tot_data(Counter(functions))


def test_marker_api_trace_json(json_data):
    data = json_data["rocprofiler-sdk-tool"]

    def get_kind_name(kind_id):
        return data.strings.buffer_records[kind_id]["kind"]

    valid_domain = (
        "MARKER_CORE_API",
        "MARKER_CONTROL_API",
        "MARKER_NAME_API",
        "MARKER_CORE_RANGE_API",
    )

    marker_data = data.buffer_records.marker_api

    for marker in marker_data:
        assert get_kind_name(marker["kind"]) in valid_domain
        assert marker["thread_id"] >= data["metadata"]["pid"]
        assert marker["end_timestamp"] >= marker["start_timestamp"]

    tot_data = Counter([m["value"] for m in data.strings.marker_api])
    check_tot_data(tot_data)


def test_perfetto_data(pftrace_data, json_data):
    import rocprofiler_sdk.tests.rocprofv3 as rocprofv3

    rocprofv3.test_perfetto_data(
        pftrace_data, json_data, ("hip", "hsa", "marker", "kernel", "memory_copy")
    )


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



#!@Python3_EXECUTABLE@

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import os
import roctx
import random

from roctx.context_decorators import RoctxRange

_prefix = ""


@RoctxRange("fib")
def fib(n, nmin):
    with RoctxRange(f"fib(n={n})" if n >= nmin else None):
        return n if n < 2 else (fib(n - 1, nmin) + fib(n - 2, nmin))


@RoctxRange("sum")
def _sum(arr):
    with RoctxRange(f"sum(nelem={len(arr)})"):
        return sum(arr)


def inefficient(n):
    roctx.rangePush(f"inefficient({n})")
    a = 0
    for i in range(n):
        a += i
        for j in range(n):
            a += j
    _len = a * n * n
    _arr = [random.random() for _ in range(_len)]
    _ret = _sum(_arr)
    roctx.rangePop()
    return _ret


def run(n):
    idx = roctx.rangeStart(f"run({n})")
    _ret_a = fib(n, max([n / 2, n - 10]))
    _ret_b = inefficient(n)
    roctx.rangeStop(idx)
    return (_ret_a, _ret_b)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-n", "--num-iterations", help="Number", type=int, default=3)
    parser.add_argument("-v", "--value", help="Starting value", type=int, default=20)
    args = parser.parse_args()

    _prefix = os.path.basename(__file__)
    roctx.mark(f"iterations: {args.num_iterations}")
    for i in range(args.num_iterations):
        with RoctxRange("main loop"):
            ans_a, ans_b = run(args.value)
            print(f"[{_prefix}] [{i}] result of run({args.value}) = {ans_a}, {ans_b}\n")



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest


def test_hsa_api_trace(hsa_trace_input_data):
    functions = []
    correlation_ids = []
    for row in hsa_trace_input_data:
        assert row["Domain"] in (
            "HSA_CORE_API",
            "HSA_AMD_EXT_API",
            "HSA_IMAGE_EXT_API",
            "HSA_FINALIZE_EXT_API",
        )
        assert int(row["Process_Id"]) > 0
        assert int(row["Thread_Id"]) >= int(row["Process_Id"])
        assert int(row["End_Timestamp"]) >= int(row["Start_Timestamp"])
        functions.append(row["Function"])
        correlation_ids.append(int(row["Correlation_Id"]))

    correlation_ids = sorted(list(set(correlation_ids)))

    # deterministic call counts
    num_queue_create_calls = 2
    num_queue_destroy_calls = 2
    num_hsa_mem_free_calls = 2

    # signal create/destroy calls
    # although the app explicitly only creates 3 signals
    # but hsa_init() internally calls hsa_signal_create
    num_hsa_signal_create_calls = 4
    num_hsa_signal_destroy_calls = 4

    # all correlation ids are unique
    assert len(correlation_ids) == len(hsa_trace_input_data)

    functions = list(functions)
    assert "hsa_shut_down" in functions
    assert functions.count("hsa_queue_create") == num_queue_create_calls
    assert functions.count("hsa_queue_destroy") == num_queue_destroy_calls
    assert functions.count("hsa_memory_free") == num_hsa_mem_free_calls
    assert functions.count("hsa_signal_create") == num_hsa_signal_create_calls
    assert functions.count("hsa_signal_destroy") == num_hsa_signal_destroy_calls


def test_kernel_trace(kernel_trace_input_data):
    valid_kernel_names = ("copyA", "copyB", "copyC")

    assert len(kernel_trace_input_data) == 3
    for row in kernel_trace_input_data:
        assert row["Kind"] == "KERNEL_DISPATCH"
        assert int(row["Agent_Id"].split(" ")[-1]) >= 0
        assert int(row["Queue_Id"]) > 0
        assert int(row["Kernel_Id"]) > 0
        assert row["Kernel_Name"] in valid_kernel_names
        assert int(row["Correlation_Id"]) > 0
        assert int(row["Workgroup_Size_X"]) == 64
        assert int(row["Workgroup_Size_Y"]) == 1
        assert int(row["Workgroup_Size_Z"]) == 1
        assert int(row["Grid_Size_X"]) == 64
        assert int(row["Grid_Size_Y"]) == 1
        assert int(row["Grid_Size_Z"]) == 1
        assert int(row["End_Timestamp"]) >= int(row["Start_Timestamp"])


def test_kernel_trace_json(json_data):
    data = json_data["rocprofiler-sdk-tool"]
    valid_kernel_names = ["copyA", "copyB", "copyC"]
    buffer_records = data["buffer_records"]
    buffer_names = data["strings"]["buffer_records"]
    kernel_dispatch_data = buffer_records["kernel_dispatch"]

    def get_kernel_name(kernel_id):
        return data["kernel_symbols"][kernel_id]["formatted_kernel_name"]

    assert len(kernel_dispatch_data) == 3

    for dispatch in kernel_dispatch_data:

        assert buffer_names[dispatch["kind"]]["kind"] == "KERNEL_DISPATCH"
        dispatch_info = dispatch["dispatch_info"]
        assert dispatch_info["agent_id"]["handle"] > 0
        assert dispatch_info["queue_id"]["handle"] > 0
        assert dispatch_info["kernel_id"] > 0

        kernel_name = get_kernel_name(dispatch_info["kernel_id"])
        assert kernel_name in valid_kernel_names

        assert dispatch["correlation_id"]["internal"] > 0
        assert dispatch_info["workgroup_size"]["x"] == 64
        assert dispatch_info["workgroup_size"]["y"] == 1
        assert dispatch_info["workgroup_size"]["z"] == 1
        assert dispatch_info["grid_size"]["x"] == 64
        assert dispatch_info["grid_size"]["y"] == 1
        assert dispatch_info["grid_size"]["z"] == 1
        assert dispatch["end_timestamp"] >= dispatch["start_timestamp"]


def test_hsa_api_trace_json(json_data):
    data = json_data["rocprofiler-sdk-tool"]
    functions = []
    correlation_ids = []

    def get_operation_name(kind_id, op_id):
        return data["strings"]["buffer_records"][kind_id]["operations"][op_id]

    def get_kind_name(kind_id):
        return data["strings"]["buffer_records"][kind_id]["kind"]

    metadata = data["metadata"]
    buffer_records = data["buffer_records"]

    valid_domain_names = (
        "HSA_CORE_API",
        "HSA_AMD_EXT_API",
        "HSA_IMAGE_EXT_API",
        "HSA_FINALIZE_EXT_API",
    )

    assert metadata["pid"] > 0
    hsa_api_data = buffer_records["hsa_api"]

    for itr in hsa_api_data:
        kind = get_kind_name(itr["kind"])
        assert kind in valid_domain_names
        assert itr["end_timestamp"] >= itr["start_timestamp"]
        functions.append(get_operation_name(itr["kind"], itr["operation"]))
        correlation_ids.append(itr["correlation_id"]["internal"])

    correlation_ids = sorted(list(set(correlation_ids)))

    # deterministic call counts
    num_queue_create_calls = 2
    num_queue_destroy_calls = 2
    num_hsa_mem_free_calls = 2

    # signal create/destroy calls
    # although the app explicitly only creates 3 signals
    # but hsa_init() internally calls hsa_signal_create
    num_hsa_signal_create_calls = 4
    num_hsa_signal_destroy_calls = 4

    # all correlation ids are unique
    assert len(correlation_ids) == len(hsa_api_data)

    functions = list(functions)
    assert "hsa_shut_down" in functions
    assert functions.count("hsa_queue_create") == num_queue_create_calls
    assert functions.count("hsa_queue_destroy") == num_queue_destroy_calls
    assert functions.count("hsa_memory_free") == num_hsa_mem_free_calls
    assert functions.count("hsa_signal_create") == num_hsa_signal_create_calls
    assert functions.count("hsa_signal_destroy") == num_hsa_signal_destroy_calls


def test_perfetto_data(pftrace_data, json_data):
    import rocprofiler_sdk.tests.rocprofv3 as rocprofv3

    rocprofv3.test_perfetto_data(
        pftrace_data, json_data, ("hip", "hsa", "marker", "kernel", "memory_copy")
    )


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)



################################################################################
# Copyright (c) 2018 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
################################################################################

import csv, sqlite3, re, sys
from functools import reduce
from txt2params import gen_params


# SQLite Database class
class SQLiteDB:
    def __init__(self, file_name):
        self.connection = sqlite3.connect(file_name)
        self.tables = {}
        self.section_index = 0

    def __del__(self):
        self.connection.close()

    # add DB table
    def add_table(self, name, descr, extra=()):
        (field_list, field_dict) = descr
        if name in self.tables:
            raise Exception('table is already added: "' + name + '"')

        # create DB table
        table_descr = []
        for field in field_list:
            table_descr.append('"%s" %s' % (field, field_dict[field]))
        for item in extra:
            table_descr.append('"%s" %s' % (item[0], item[1]))
        stm = "CREATE TABLE " + name + " (%s)" % ", ".join(table_descr)
        cursor = self.connection.cursor()
        cursor.execute(stm)
        self.connection.commit()

        # register table
        fields_str = ",".join(map(lambda x: '"' + x + '"', field_list))
        templ_str = ",".join("?" * len(field_list))
        stm = "INSERT INTO " + name + "(" + fields_str + ") VALUES(" + templ_str + ");"
        self.tables[name] = stm

        return (cursor, stm)

    # add columns to table
    def add_columns(self, name, columns):
        cursor = self.connection.cursor()
        for item in columns:
            stm = "ALTER TABLE " + name + ' ADD COLUMN "%s" %s' % (item[0], item[1])
            cursor.execute(stm)
        self.connection.commit()

    # add columns with expression
    def add_data_column(self, table_name, data_label, data_type, data_expr):
        cursor = self.connection.cursor()
        cursor.execute(
            'ALTER TABLE %s ADD COLUMN "%s" %s' % (table_name, data_label, data_type)
        )
        cursor.execute("UPDATE %s SET %s = (%s);" % (table_name, data_label, data_expr))

    def change_rec_name(self, table_name, rec_id, rec_name):
        self.connection.execute(
            "UPDATE " + table_name + ' SET Name = ? WHERE "Index" = ?', (rec_name, rec_id)
        )

    def change_rec_tid(self, table_name, rec_id, tid):
        self.connection.execute(
            "UPDATE " + table_name + ' SET tid = ? WHERE "Index" = ?', (tid, rec_id)
        )

    def change_rec_fld(self, table_name, fld_expr, rec_pat):
        self.connection.execute(
            "UPDATE " + table_name + " SET " + fld_expr + " WHERE " + rec_pat
        )

    def table_get_record(self, table_name, rec_pat):
        cursor = self.connection.execute(
            "SELECT * FROM " + table_name + " WHERE " + rec_pat
        )
        raws = cursor.fetchall()
        if len(raws) != 1:
            raise Exception(
                "Record (" + rec_pat + ') is not unique, table "' + table_name + '"'
            )
        return list(raws[0])

    # populate DB table entry
    def insert_entry(self, table, val_list):
        (cursor, stm) = table
        cursor.execute(stm, val_list)

    # populate DB table entry
    def commit_entry(self, table, val_list):
        self.insert_entry(table, val_list)
        self.connection.commit()

    # populate DB table data
    def insert_table(self, table, reader):
        for val_list in reader:
            if not val_list[-1]:
                val_list.pop()
            self.insert_entry(table, val_list)
        self.connection.commit()

    # return table fields list
    def _get_fields(self, table_name):
        cursor = self.connection.execute("SELECT * FROM " + table_name)
        return list(map(lambda x: '"%s"' % (x[0]), cursor.description))

    # return table raws list
    def _get_raws(self, table_name):
        cursor = self.connection.execute("SELECT * FROM " + table_name)
        return cursor.fetchall()

    def _get_raws_indexed(self, table_name):
        cursor = self.connection.execute(
            "SELECT * FROM " + table_name + ' order by "Index" asc;'
        )
        return cursor.fetchall()

    def _get_raw_by_id(self, table_name, rec_id):
        cursor = self.connection.execute(
            "SELECT * FROM " + table_name + ' WHERE "Index"=?', (rec_id,)
        )
        raws = cursor.fetchall()
        if len(raws) != 1:
            raise Exception('Index is not unique, table "' + table_name + '"')
        return list(raws[0])

    def table_get_raws(self, table_name):
        return self._get_raws(table_name)

    # dump CSV table
    def dump_csv(self, table_name, file_name):
        if not re.search(r"\.csv$", file_name):
            raise Exception('wrong output file type: "' + file_name + '"')

        fields = self._get_fields(table_name)
        with open(file_name, mode="w") as fd:
            fd.write(",".join(fields) + "\n")
            for raw in self._get_raws(table_name):
                tmp = list(raw)
                for idx in range(len(tmp)):
                    if type(tmp[idx]) == str:
                        if not (tmp[idx][0] == tmp[idx][-1] == '"'):
                            tmp[idx] = '"' + tmp[idx] + '"'
                raw = tuple(tmp)
                fd.write(reduce(lambda a, b: str(a) + "," + str(b), raw) + "\n")

    # dump JSON trace
    def open_json(self, file_name):
        if not re.search(r"\.json$", file_name):
            raise Exception('wrong output file type: "' + file_name + '"')
        with open(file_name, mode="w") as fd:
            fd.write('{ "traceEvents":[{}\n')

    def close_json(self, file_name):
        if not re.search(r"\.json$", file_name):
            raise Exception('wrong output file type: "' + file_name + '"')
        with open(file_name, mode="a") as fd:
            fd.write("}")

    def label_json(self, pid, label, file_name):
        if not re.search(r"\.json$", file_name):
            raise Exception('wrong output file type: "' + file_name + '"')
        with open(file_name, mode="a") as fd:
            fd.write(
                ',{"args":{"name":"%s"},"ph":"M","pid":%s,"name":"process_name","sort_index":%d}\n'
                % (label, pid, self.section_index)
            )
        self.section_index += 1

    def flow_json(
        self, base_id, from_pid, from_us_list, to_pid, to_us_dict, corr_id_list, file_name
    ):
        if not re.search(r"\.json$", file_name):
            raise Exception('wrong output file type: "' + file_name + '"')
        with open(file_name, mode="a") as fd:
            dep_id = base_id
            for ind in range(len(from_us_list)):
                corr_id = corr_id_list[ind] if (len(corr_id_list) != 0) else ind
                if corr_id in to_us_dict:
                    (from_ts, stream_id, tid) = from_us_list[ind]
                    to_ts = to_us_dict[corr_id]
                    if from_ts > to_ts:
                        from_ts = to_ts
                    fd.write(
                        ',{"ts":%d,"ph":"s","cat":"DataFlow","id":%d,"pid":%d,"tid":%d,"name":"dep"}\n'
                        % (from_ts, dep_id, from_pid, tid)
                    )
                    fd.write(
                        ',{"ts":%d,"ph":"t","cat":"DataFlow","id":%d,"pid":%d,"tid":%d,"name":"dep"}\n'
                        % (to_ts, dep_id, to_pid, stream_id)
                    )
                    dep_id += 1

    def metadata_json(self, jsonfile, sysinfo_file):
        params = gen_params(sysinfo_file)
        with open(jsonfile, mode="a") as fd:
            cnt = 0
            fd.write("],\n")
            fd.write('"otherData": {\n')
            for nkey in sorted(params.keys()):
                key = nkey[1]
                cnt = cnt + 1
                if cnt == len(params):
                    fd.write('    "' + key + '": "' + params[nkey] + '"\n')
                else:
                    fd.write('    "' + key + '": "' + params[nkey] + '",\n')
            fd.write("  }\n")

    def dump_json(self, table_name, data_name, file_name):
        if not re.search(r"\.json$", file_name):
            raise Exception('wrong output file type: "' + file_name + '"')

        sub_ptrn = re.compile(r'(^"|"$)')
        name_ptrn = re.compile(r"(name|Name)")

        table_fields = self._get_fields(table_name)
        table_raws = self._get_raws(table_name)
        data_fields = self._get_fields(data_name)
        data_raws = self._get_raws(data_name)

        with open(file_name, mode="a") as fd:
            table_raws_len = len(table_raws)
            for raw_index in range(table_raws_len):
                if (raw_index == table_raws_len - 1) or (raw_index % 1000 == 0):
                    sys.stdout.write(
                        "\rdump json "
                        + str(raw_index)
                        + ":"
                        + str(len(table_raws))
                        + " " * 100
                    )

                vals_list = []
                values = list(table_raws[raw_index])
                for value_index in range(len(values)):
                    label = table_fields[value_index]
                    value = values[value_index]
                    if name_ptrn.search(label):
                        value = sub_ptrn.sub(r"", value)
                    if label != '"Index"':
                        if label == '"dur"' and value == 0:
                            vals_list.append('%s:"%s"' % (label, "1"))
                        elif label == '"pid"' or label == '"tid"':
                            vals_list.append('%s:%d' % (label, value))
                        else:
                            vals_list.append('%s:"%s"' % (label, value))

                args_list = []
                data = list(data_raws[raw_index])
                for value_index in range(len(data)):
                    label = data_fields[value_index]
                    value = data[value_index]
                    if label[:3] == '"__':
                        continue
                    if name_ptrn.search(label):
                        value = sub_ptrn.sub(r"", value)
                    if label != '"Index"' and label != '"roctx-range"':
                        args_list.append('%s:"%s"' % (label, value))

                fd.write(
                    ',{"ph":"%s",%s,\n  "args":{\n    %s\n  }\n}\n'
                    % ("X", ",".join(vals_list), ",\n    ".join(args_list))
                )

        sys.stdout.write("\n")

    # execute query on DB
    def execute(self, cmd):
        cursor = self.connection.cursor()
        cursor.execute(cmd)

    # commit DB
    def commit(self):
        self.connection.commit()

    # close DB
    def close(self):
        self.connection.commit()
        self.connection.close()

    # access DB
    def get_raws(self, table_name):
        cur = self.connection.cursor()
        cur.execute("SELECT * FROM %s" % table_name)
        return cur.fetchall()

    # return CSV descriptor
    # list of fields and dictionaly for the fields types
    def _get_csv_descr(self, table_name, fd):
        reader = csv.DictReader(fd)
        field_names = reader.fieldnames
        if not field_names[-1]:
            field_names.pop()
        field_types = {}

        for entry in reader:
            fields_left = [f for f in field_names if f not in field_types.keys()]
            # all fields processed
            if not fields_left:
                break

            for field in fields_left:
                data = entry[field]
                # need data for the field to be processed
                if len(data) == 0:
                    continue

                if data.isdigit():
                    field_types[field] = "INTEGER"
                else:
                    field_types[field] = "TEXT"

        if len(fields_left) > 0:
            raise Exception("types not found for fields: ", fields_left)
        return (field_names, field_types)

    # add CSV table
    def add_csv_table(self, table_name, file_name, extra=()):
        with open(file_name, mode="r") as fd:
            # get CSV table descriptor
            descr = self._get_csv_descr(table_name, fd)
            # reader to populate the table
            fd.seek(0)
            reader = csv.reader(fd)
            reader.next()
            table = self.add_table(table_name, descr, extra)
            self.insert_table(table, reader)


##############################################################################################



# MIT License
#
# Copyright (c) 2023-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import absolute_import

import otf2
import pandas as pd

from otf2.events import Enter, Leave


class Region(object):
    """ """

    def __init__(self, _enter, _leave, _depth, _location=None):
        if _enter.region != _leave.region:
            _location_info = f"\n{_location}" if _location else ""
            raise ValueError(
                f"enter region != leave region :: '{_enter}' != '{_leave}'{_location_info}"
            )

        if _depth < 0:
            _location_info = f". location: '{_location}'" if _location else ""
            raise ValueError(
                f"negative depth ({_depth})! enter: '{_enter}'. leave: '{_leave}'{_location_info}"
            )

        self.region = _enter.region
        self.depth = _depth
        self.name = _enter.region.name
        self.attributes = [
            itr for itr in [_enter.attributes, _leave.attributes] if itr is not None
        ]
        self.enter_nsec = _enter.time
        self.leave_nsec = _leave.time
        self.delta_nsec = _leave.time - _enter.time
        if self.delta_nsec < 0:
            raise ValueError(
                f"negative timestamp delta :: '{_enter.time}' > '{_leave.time}'"
            )

        for itr in self.attributes:
            for key, val in itr.items():
                _key = f"{key.name}"
                if not hasattr(self, _key):
                    setattr(self, _key, val)

        if not hasattr(self, "category"):
            self.category = "unk"

    def __str__(self):
        return f"{self.name:<35} :: {self.delta_nsec} nsec"


class OTF2Reader:
    """Read in perfetto protobuf output"""

    def __init__(self, filename):
        self.filename = filename if isinstance(filename, (list, tuple)) else [filename]

    def read(self):
        def _read_trace(trace_name):
            trace = otf2.reader.Reader(trace_name)
            # print(f"Read {len(trace.definitions.strings)} string definitions")
            # for string in trace.definitions.strings:
            #     print(f"String definition with value '{string}' in trace.")
            # print("Read {} events".format(len(trace.events)))

            events = [[loc, evt] for loc, evt in trace.events]
            locations = [itr for itr in trace.definitions.locations]
            location_groups = [itr for itr in trace.definitions.location_groups]
            system_tree_nodes = [itr for itr in trace.definitions.system_tree_nodes]

            call_stack = {}
            partial_call_stack = {}

            for itr in system_tree_nodes:
                call_stack[itr] = {}
                partial_call_stack[itr] = {}

            for itr in location_groups:
                call_stack[itr.system_tree_parent][itr] = {}
                partial_call_stack[itr.system_tree_parent][itr] = {}

            for itr in locations:
                call_stack[itr.group.system_tree_parent][itr.group][itr] = []
                partial_call_stack[itr.group.system_tree_parent][itr.group][itr] = []

            for location, event in events:
                _stree = location.group.system_tree_parent
                _group = location.group
                _partial = partial_call_stack[_stree][_group][location]
                if isinstance(event, Enter):
                    # expected length
                    _elen = len(_partial) + 1
                    _partial.append(event)
                elif isinstance(event, Leave):
                    # expected length
                    _elen = len(_partial) - 1
                    _depth = len(_partial)
                    _leave = event

                    # it appears that on MI300, the end of A may exceed the
                    # begin of B kernels very slightly (i.e. overlap in same
                    # stream/queue). This leads to slightly out of order
                    # Enter/Leave regions and thus we need to occasionally
                    # search further back in the callstack to find the correct
                    # Enter region
                    _enter = _partial[-1]
                    if _enter.region == _leave.region:
                        _partial.pop()
                    else:
                        for ridx, ritr in enumerate(reversed(_partial)):
                            if ritr.region == _leave.region:
                                _enter = _partial.pop(len(_partial) - ridx - 1)
                                break

                    # below is what is expected on non-MI300
                    # _enter = _partial.pop()

                    # add the region
                    call_stack[_stree][_group][location].append(
                        Region(_enter, _leave, _depth - 1, location)
                    )

                # modified length
                _mlen = len(partial_call_stack[_stree][_group][location])
                # if modified length != expected length
                if _mlen != _elen:
                    raise RuntimeError(
                        f"Modified length ({_mlen}) != Expected length({_elen}) for {event} at {location}"
                    )

            data = {
                "system_tree_node": [],
                "location_group": [],
                "location": [],
                "region": [],
                "attributes": [],
                "depth": [],
                "name": [],
                "category": [],
                "start_ts": [],
                "end_ts": [],
            }

            for tree, lgitr in call_stack.items():
                for group, gitr in lgitr.items():
                    for loc, ritr in gitr.items():
                        for region in ritr:
                            data["system_tree_node"] += [tree]
                            data["location_group"] += [group]
                            data["location"] += [loc]
                            data["region"] += [region.region]
                            data["attributes"] += [region.attributes]
                            data["depth"] += [region.depth]
                            data["category"] += [region.category]
                            data["name"] += [region.name]
                            data["start_ts"] += [region.enter_nsec]
                            data["end_ts"] += [region.leave_nsec]

            return (trace, pd.DataFrame.from_dict(data))

        readers = []
        df = pd.DataFrame()
        for itr in self.filename:
            _reader, _df = _read_trace(itr)
            readers += [_reader]
            df = pd.concat([df, _df])

        return (df, readers)


def read_trace(filename):
    data = OTF2Reader(filename).read()[0]

    print(f"\nDATA:\n{data}")

    attributes = list(data["attributes"])

    print(f"\nATTRIBUTES:\n{attributes}")



#!/usr/bin/env python3

# MIT License
#
# Copyright (c) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import pytest


def test_marker_api_trace(json_data):
    data = json_data["rocprofiler-sdk-tool"]

    def get_kind_name(kind_id):
        return data["strings"]["buffer_records"][kind_id]["kind"]

    def get_region_name(corr_id):
        for itr in data["strings"]["marker_api"]:
            if itr.key == corr_id:
                return itr.value
        return None

    valid_domain = (
        "MARKER_CORE_API",
        "MARKER_CONTROL_API",
        "MARKER_NAME_API",
        "MARKER_CORE_RANGE_API",
    )

    buffer_records = data["buffer_records"]
    marker_data = buffer_records["marker_api"]
    tot_data = {}
    thr_data = {}
    for marker in marker_data:
        assert get_kind_name(marker["kind"]) in valid_domain
        assert marker.thread_id >= data["metadata"]["pid"]
        assert marker.end_timestamp >= marker.start_timestamp

        if marker.thread_id not in thr_data.keys():
            thr_data[marker.thread_id] = {}

        corr_id = marker.correlation_id.internal
        assert corr_id > 0, f"{marker}"
        name = get_region_name(corr_id)
        if not name.startswith("roctracer/roctx"):
            assert "run" in name, f"{marker}"
            if name not in thr_data[marker.thread_id].keys():
                thr_data[marker.thread_id][name] = 1
            else:
                thr_data[marker.thread_id][name] += 1

        if name not in tot_data.keys():
            tot_data[name] = 1
        else:
            tot_data[name] += 1

    assert tot_data["roctracer/roctx v4.1"] == 1
    assert tot_data["run"] == 2
    assert tot_data["run/iteration"] == 1000
    assert tot_data["run/iteration/sync"] == 100
    assert tot_data["run/rank-0/thread-0/device-0/begin"] == 1
    assert tot_data["run/rank-0/thread-0/device-0/end"] == 1
    assert len(tot_data.keys()) >= 8

    for tid, titr in thr_data.items():
        assert titr["run"] == 1
        assert titr["run/iteration"] == 500
        assert titr["run/iteration/sync"] == 50
        assert len(titr.keys()) >= 5


def test_perfetto_data(pftrace_data, json_data):
    import rocprofiler_sdk.tests.rocprofv3 as rocprofv3

    rocprofv3.test_perfetto_data(pftrace_data, json_data, ("memory_copy", "marker"))


def test_otf2_data(otf2_data, json_data):
    import rocprofiler_sdk.tests.rocprofv3 as rocprofv3

    rocprofv3.test_otf2_data(otf2_data, json_data, ("memory_copy", "marker"))


if __name__ == "__main__":
    exit_code = pytest.main(["-x", __file__] + sys.argv[1:])
    sys.exit(exit_code)
